//These functions comprise the Metropolis-Hastings updates utilized 
//in the hrm.cpp

//Originally written by Matt Johnson (1999)
//Updated by Lou Mariano (2000, 2001, 2002)

#include<stdlib.h>
#include<math.h>
#include<iostream.h>
#include "PCM.hpp"

double genrand_real3();                 //***RNG***
#define drnunf() genrand_real3()        //***RNG***

#define drnnof() rnorm()




void rdisc(int nr,int imin,int nmass,double *prob,int *ir){
  for(int j=0;j<nr;j++){
    ir[j]=imin;
    double r=drnunf();
    int i=0;
    while(r>prob[i]){
      ++ir[j];
      ++i;
    }
    return;
  }
}

double lognorm_dens(double x, double mean, double sd);
double rgamma(double);
double rnorm(void);
double prob_calc(Quest task,double theta,int x);
void calc_trans(int ncat,double mu,double sigma,Mat<double> &trans );
void alt_calc_trans(int ncat,double mu,double sigma,Mat<double> &trans);
void pcm (Quest task,double theta,double prob[]);
void cumsum(int dim,double gamma[],double eta[]);

//#########################################################################

void update_beta_cons(int nresp,Quest &task,Quest &taskc
		 ,double theta[],int x[],int xc[]){
  double prob[task.ncat],cprob[task.ncat],probc[task.ncat];
  double cprobc[task.ncat],lprior,clprior,lrat,u;
  double llike=0,cllike=0,acc=0;
  Quest cand=task;
  Quest candc=taskc;
  double t;
  t=drnnof();
  double cvar=.15;
  cand.SetBeta(task.beta+cvar*t);
  candc.SetBeta(taskc.beta-cvar*t);

  for(int i=0;i<nresp;i++){
    if(x[i]==task.ncat){
      llike+=0;
      cllike+=0;
    }
    else{
      pcm(task,theta[i],prob);
      pcm(taskc,theta[i],probc);
      pcm(cand,theta[i],cprob);
      pcm(candc,theta[i],cprobc);
      llike += log(prob[x[i]])+log(probc[xc[i]]);
      cllike += log(cprob[x[i]])+log(cprobc[xc[i]]);
    }
  }
  lprior=0-.5*pow(task.beta,2)/10;
  clprior=0-.5*pow(cand.beta,2)/10;
  lrat=cllike+clprior-lprior-llike;
  u=drnunf();
  if(u < exp(lrat)){
    task=cand;
    taskc=candc;
  }
  return;
}


//#######################################################################

void update_gamma(int nresp,Quest &task,double theta[],int x[]){
  double prob[task.ncat],cprob[task.ncat];
  double lprior=0,clprior=0,lrat=0,u;
  double llike=0,cllike=0;
  double t;
  Quest cand;
  
  cand.SetNcat(task.ncat);
  cand.SetBeta(task.beta);
  cand.SetGamma(task.gamma);
  
  for(int j=0;j<task.ncat-2;j++){
    t=drnnof();
    double cvar=.2;
    cand.gamma[j] = task.gamma[j]+cvar*t;
    cand.gamma[task.ncat-2] = task.gamma[task.ncat-2]-cvar*t;
    for(int i=0;i<nresp;i++){
      if(x[i]==task.ncat){
	llike+=0;
	cllike+=0;
      }
      else{
	pcm(task,theta[i],prob);
	pcm(cand,theta[i],cprob);
	llike += log(prob[x[i]]);
	cllike += log(cprob[x[i]]);
      }
    }
    lprior=0-.5*pow(task.gamma[j],2)/3;
    clprior=0-.5*pow(cand.gamma[j],2)/3;
    lrat=cllike+clprior-llike-lprior;
    
    u=drnunf();
    if(u < exp(lrat)){
      task.gamma[j]=cand.gamma[j];
      task.gamma[task.ncat-2]=cand.gamma[task.ncat-2];
    }
  }
  cand.FreeNcat();
  return;
}
  
		
//#################################################################

void update_lat_resp(Quest task,double theta,Mat<double> tran[],
		     int sc[],int nrate,int &xi){
  double prob[task.ncat],cprob[task.ncat+1],r[task.ncat];
  double sum=0,z;
  double phi,norm[nrate][task.ncat];
  int nr=1,imin=0,nmass=task.ncat;
  int ir[nr];
  pcm(task,theta,prob);
  
  for(int r=0;r<nrate;r++){
    for(int k=0;k<task.ncat;k++){
      norm[r][k]=0;
    }}
  for(int r=0;r<nrate;r++){
    for(int k=0;k<task.ncat;k++){
      for(int j=0;j<task.ncat;j++){
	norm[r][k] += tran[r].GetElem(k,j);
      }
    }
  }
  

  for(int i=0;i<task.ncat;i++){
    r[i]=1;
    for(int j=0;j<nrate;j++){
      phi= sc[j]==task.ncat ? 1 : 
	tran[j].GetElem(i,sc[j])/norm[j][i];
      r[i] *=phi;
      // r[i] is the prob. P(sc[j] | xi =i);
    }
    r[i] *= prob[i];
    sum += r[i];
  }
  cprob[task.ncat]=0;
  cprob[task.ncat-1]=1;
  cprob[0]=r[0]/sum;
  for(int i=1;i< task.ncat -1;i++){
    cprob[i]=cprob[i-1]+r[i]/sum;
  }
  rdisc(nr,imin,nmass,cprob,ir);
  xi= ir[0];
}
  
  
//#####################################################################


void update_rater(Quest task[],int nresp,int nquest,
		  Mat<int> x,Mat<int> xis,double &shift,double &scale,
		  Mat<double> &trans,double cshvar,double cscvar){
  double cshift,llike_diff=0;
  double cphi,phi;
  cshift=shift+cshvar*drnnof();
  Mat<double> ctrans;
  int lat_resp;
  int score,maxcat=0;

  for(int j=0;j<nquest;j++)
    maxcat = task[j].ncat > maxcat ? task[j].ncat : maxcat;
  ctrans.Mat2(maxcat,maxcat);
  double norm[maxcat],cnorm[maxcat],tmp[maxcat],tmpc[maxcat];
  
  calc_trans(maxcat,cshift,scale,ctrans);
  
  for(int j=0;j<nquest;j++){
    for(int i=0;i<nresp;i++){
      score=x.GetElem(i,j);
      if(score<task[j].ncat){
	lat_resp=xis.GetElem(i,j);
	trans.GetRow(lat_resp,tmp);
	ctrans.GetRow(lat_resp,tmpc);
	cumsum(task[j].ncat,tmp,norm);
	cumsum(task[j].ncat,tmpc,cnorm);
      	phi = log(tmp[score]/norm[task[j].ncat-1]);
	cphi= log(tmpc[score]/cnorm[task[j].ncat-1]);
	llike_diff += cphi-phi;
      }
    }
  }

  llike_diff += .5*(pow(shift,2)/10.)-.5*(pow(cshift,2)/10.);

  if(drnunf()<exp(llike_diff)){
    shift=cshift;
    for(int k=0;k<maxcat;k++)
      for(int l=0;l<maxcat;l++)
	trans.SetElem(k,l,ctrans.GetElem(k,l));
  }
  

  llike_diff=0;
  
  double cscale;
  cscale=scale*exp(cscvar*drnnof());

  calc_trans(maxcat,shift,cscale,ctrans);
  
  for(int j=0;j<nquest;j++){
    for(int i=0;i<nresp;i++){
      score=x.GetElem(i,j);
      if(score<task[j].ncat){
	lat_resp=xis.GetElem(i,j);
	trans.GetRow(lat_resp,tmp);
	ctrans.GetRow(lat_resp,tmpc);
	cumsum(task[j].ncat,tmp,norm);
	cumsum(task[j].ncat,tmpc,cnorm);
	phi = log(tmp[score]/norm[task[j].ncat-1]);
	cphi= log(tmpc[score]/cnorm[task[j].ncat-1]);
	llike_diff += cphi-phi;
      }
    }
  }

  llike_diff += .5*(pow(log(scale),2)/10)-.5*(pow(log(cscale),2)/10) - log(cscale) + log(scale);
  llike_diff += log( lognorm_dens(scale,log(cscale),cscvar) ) - log( lognorm_dens(cscale,log(scale),cscvar) );
  if(drnunf()<exp(llike_diff)){
    scale=cscale;
    for(int k=0;k<maxcat;k++)
      for(int l=0;l<maxcat;l++)
	trans.SetElem(k,l,ctrans.GetElem(k,l));
  }
  ctrans.FreeMat2();

  return;
  
}


//#################################################################

void calc_trans(int ncat,double mu,double sigma,Mat<double> &trans){
  double z;
  double q,p;
  double sum;//to verify row nonzero row-sums
  bool zerosum=false;//is there a zero sum row in trans  
  
  for(int l=0;l<ncat;l++){//cycle through ideal scores
    sum =0;//initialize row-sum to zero
    for(int k=0;k<ncat;k++){//cycle through categories
      z=(k-l-mu)/sigma;
      q=0-pow(z,2)/2;
      p=exp(q);
      trans.SetElem(l,k,p);
      sum += p;
    }
    if (sum == 0){
      zerosum=true;
    }
  }    

  if (zerosum){
  alt_calc_trans(ncat, mu, sigma, trans);
  }

}


void alt_calc_trans(int ncat,double mu,double sigma,Mat<double> &trans){

  double z[ncat];
  double q[ncat],p[ncat];


  for(int l=0;l<ncat;l++){//cycle through ideal scores
    for(int k=0;k<ncat;k++){//cycle through categories
      z[k]=(k-l-mu);
    }
    for(int k=0;k<ncat;k++){//cycle through categories
      p[k]=0;
      for(int h=0;h<ncat;h++){
	q[h]= pow(z[h],2) - pow(z[k],2);
	p[k]+= exp( -.5 * q[h] / pow(sigma,2) );
      }
      trans.SetElem(l,k,1/p[k]);
    }
  }
}

//####################################################################

void update_var(int nresp, double theta[], double &mu,double &tau){
  double z,alpha=1,beta=1,xbar=0,sd,mn,pvar=10;
  int nr=1;
  double ss=0,scale,chi[nr];
  for(int j=0;j<nresp;j++){
    xbar+=theta[j]/nresp;
    ss += pow(theta[j]-mu,2);
  }
  alpha += nresp/2.;
  beta += ss/2.;
  chi[0]=rgamma(alpha);
  tau=beta/(chi[0]);
  z=drnnof();
  mn=nresp*pvar*xbar/(nresp*pvar+tau);
  sd=sqrt(tau*pvar/(nresp*pvar+tau));
  mu=mn+z*sd;
  return;
}

//###################################################################
  
//###################################################################
void update_theta(int nquest,Quest task[],double &theta,
		  double mu,double tau,int x[]){

  
  double prob,cprob;
  double cth,u;
  u=drnnof();
  double cvar=1.5;
  cth=theta+cvar*u;
  double llike=0,lprior;
  double cllike=0,clprior,lrat;
  for(int j=0;j<nquest;j++){
    if(x[j]==task[j].ncat){
      llike+=0;
      cllike+=0;
    }
    else{
      prob=prob_calc(task[j],theta,x[j]); 
      // Calculate probability for old theta     
      cprob=prob_calc(task[j],cth,x[j]);  // Calcuate prob for candidate theta
      llike += log(prob);                 // Log-Like for old theta
      cllike += log(cprob);               // Log-Like for cand theta
    }
  }
  
  lprior= 0-.5*pow(theta-mu,2)/tau;
  clprior= 0-.5*pow(cth-mu,2)/tau;
  lrat=cllike+clprior-llike-lprior;
  u=drnunf();
  if(u < exp(lrat)){
    theta=cth;
  }
  return;
}

//##############################################################3

double prob_calc(Quest task,double theta,int x){
  double prob[task.ncat];
  pcm(task,theta,prob);
  return(prob[x]);
}

//##############################################################3

double lognorm_dens(double x, double mean, double sd){
  //This function will calculate the density under a log-normal pdf.
  double hold, var;

  var = pow(sd,2);
  hold = -.5 * pow( ( log(x) - mean ) , 2) / var;
  hold = exp(hold);

  const double PI=3.14159265358979323;
  const double marg=sqrt( 2 * PI );

  hold = hold/(x * marg * sd);

  return(hold);
}
